# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

from event import Signal

class Selection(object):
    """Represents a selection on a Graph object.

    Translates between frame numbers and pixels.

    """
    def __init__(self, graph, cursor):
        self._graph = graph
        self._cursor = cursor
        self.changed = Signal()
        self.unselect()
        self._graph.changed.connect(self._update)

    def _update(self):
        "Called when self._graph changes."
        self.changed()

    def set(self, start, end):
        "Set selection bounds (in frames)"
        self.start = start
        self.end = end
        self._cursor.set_frame(min(self.start, self.end))
        self.changed()

    def unselect(self):
        self.set(0, 0)

    def select_all(self):
        self.set(0, self._graph.numframes())

    def selected(self):
        return self.start != self.end

    def pin(self, pixel):
        "The pixel is an index in the graph."
        start = self._graph.pxltofrm(pixel)
        self.set(start, start)

    def extend(self, pixel):
        "The pixel is an index in the graph."
        self.end = self._graph.pxltofrm(pixel)
        self._cursor.set_frame(min(self.start, self.end))
        self.changed()

    def pixels(self):
        """Returns pixel position for selection: `(start, end)`.

        `start` is always lower than or equals to `end`.

        """
        frame_start, frame_end = self.get()
        pix_start = self._graph.frmtopxl(frame_start)
        pix_end = self._graph.frmtopxl(frame_end)
        return pix_start, pix_end

    def get(self):
        """Returns frames index for selection: (start, end).

        start is always lower than or equal to end.

        """
        start = self.start
        end = self.end
        if start > end:
            start, end = end, start
        return start, end


def test_selection():
    from graphmodel import Graph
    from mock import Fake

    class FakeGraph():
        density = 10
        numframes = (lambda(self): 5000)
        _view_start = 100.1
        _view_end = 200.1
        changed = Fake()

        def frmtopxl(self, f):
            return int(round(f - self._view_start) / self.density)

        def pxltofrm(self, p):
            return int(round(self._view_start + p * self.density))

    x1, x2 = 10, 100
    selection = Selection(FakeGraph(), Fake())
    selection.pin(x1)
    selection.extend(x2)
    selection._update()
    assert selection.pixels() == (10, 100)
    assert selection.get() == (100 + 10 * x1, 100 + 10 * x2)

    # invert selection order
    selection.pin(x2)
    selection.extend(x1)
    selection._update()
    assert selection.pixels() == (10, 100)
    assert selection.get() == (100 + 10 * x1, 100 + 10 * x2)

    # select all
    selection.select_all()
    assert selection.get() == (0, 5000)
    assert selection.selected() == True

    # unselect
    selection.unselect()
    assert selection.selected() == False

if __name__ == "__main__":
    test_selection()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import gtk
import cairo

# -- Base classes for drawing sound visualization.
#
# CairoWidget, LayeredCairoWidget, and LayeredGraphView are defined as
# successive subclasses only for code clarity. Everything could as
# well be stuck in LayeredGraphView.
#
class CairoWidget(gtk.DrawingArea):

    __gsignals__ = {"expose-event": "override"}

    def __init__(self):
        super(CairoWidget, self).__init__()
    
    def do_expose_event(self, event):
        context = self.window.cairo_create()
        context.rectangle(event.area.x, event.area.y,
                          event.area.width, event.area.height)
        context.clip()
        width, height = self.window.get_size()
        self.draw(context, width, height)

    def redraw(self):
        # queue_draw() emits an expose event. Double buffering is used
        # automatically in the expose event handler.
        self.queue_draw()

    def draw(self, context, width, height):
        """Must be overriden to draw to the cairo context."""
        pass


class LayeredCairoWidget(CairoWidget):
    """A widget with several layers.

    This widget paints itself by successively passing its context to
    layer objects. The draw() method of a layer object must paint to
    the context.
    
    """
    def __init__(self):
        super(LayeredCairoWidget, self).__init__()
        self.layers = []
        
    def draw(self, context, width, height):
        for layer in self.layers:
            layer.draw(context, width, height)


class LayeredGraphView(LayeredCairoWidget):
    """A layered widget dedicated to a Graph object.
    
    Every time the widget is resized, the new width is passed to the
    Graph object.

    """
    def __init__(self, graph):
        super(LayeredGraphView, self).__init__()
        self._graph = graph
        self.connect("size_allocate", self.resized)

    def resized(self, widget, rect):
        self._graph.set_width(rect.width)


# -- The sound visualization widget, composed of several layers:
#
#    * waveform
#    * selection
#    * cursor
#
class GraphView(LayeredGraphView):
    """Sound visualization widget for the main window.

    * Three graphical layers: waveform, selection, and cursor.
    * Mouse event listeners act on models (scroll and selection).

    """
    def __init__(self, graph, selection, cursor):
        super(GraphView, self).__init__(graph)
        self._graph = graph
        self._selection = selection
        self.layers.append(BackgroundLayer(self, selection))
        self.layers.append(WaveformLayer(self, graph))
        self.layers.append(SelectionLayer(self, selection))
        self.layers.append(CursorLayer(self, cursor))
        MouseSelection(self, selection)
        MouseScroll(self, graph)
        MouseMiddleClick(self, graph)
        PointerStyle(self, selection)


# -- Layers that can be added to LayeredGraphview.
#
class WaveformLayer(object):
    """A layer for LayeredGraphView.

    It paints the graph (the waveform). The cairo surface is cached,
    it is redrawn only when graph has changed.

    """
    def __init__(self, layered, graph):
        self._layered = layered
        self._graph = graph
        self._cache = None
        self.wavecolor = 0.0, 0.47058823529411764, 1.0
        graph.changed.connect(self.update)

    def update(self):
        self._cache = None
        self._layered.redraw()

    def draw_channel(self, values, surface, width, height):
        "Draw one sound channel on a cairo surface."
        c = cairo.Context(surface)

        # Line at zero
        c.set_line_width(1)
        c.set_source_rgb(0.2, 0.2, 0.2)
        c.move_to(0, round(height / 2) + 0.5)
        c.line_to(width, round(height / 2) + 0.5)
        c.stroke()

        # Waveform
        c.set_source_rgb(*self.wavecolor)
        for i, (mini, maxi) in enumerate(values):
            # -1 <= mini <= maxi <= 1
            # 0 <= ymin <= ymax <= height - 1
            x = i
            ymin = round((-mini * 0.5 + 0.5) * (height - 1))
            ymax = round((-maxi * 0.5 + 0.5) * (height - 1))
            if ymin == ymax:
                # Fill one pixel 
                c.rectangle(x, ymin, 1, 1)
                c.fill()
            else:
                # Draw a line from min to max
                c.move_to(x + 0.5, ymin)
                c.line_to(x + 0.5, ymax)
                c.stroke()
        
    def draw(self, context, width, height):
        "Draw all sound channels."
        if not self._cache:
            surface = context.get_target()
            self._cache = surface.create_similar(cairo.CONTENT_COLOR_ALPHA,
                                                 width, height)
            c = cairo.Context(self._cache)
            channels = self._graph.channels()
            numchan = len(channels)
            for i in range(numchan):
                s = surface.create_similar(cairo.CONTENT_COLOR_ALPHA,
                                           width, height / numchan)
                self.draw_channel(channels[i], s, width, height / numchan)
                c.set_source_surface(s, 0, (height / numchan) * i)
                c.set_operator(cairo.OPERATOR_OVER)
                c.paint()
                
        context.set_source_surface(self._cache, 0, 0)
        context.set_operator(cairo.OPERATOR_OVER)
        context.paint()


class BackgroundLayer(object):
    """A layer for LayeredGraphView.

    """
    def __init__(self, layered, selection):
        self._layered = layered
        self._cache = None
        self._selection = selection
        self._selection.changed.connect(self.update)

    def update(self):
        self._cache = None
        self._layered.redraw()

    def draw(self, context, width, height):
        if not self._cache:
            surface = context.get_target()
            self._cache = surface.create_similar(cairo.CONTENT_COLOR,
                                                 width, height)
            c = cairo.Context(self._cache)

            # Black background
            c.set_source_rgb(0, 0, 0)
            c.paint()

            # Selection background
            if self._selection.selected():
                c.set_source_rgb(0, 0, 0.2)
                start, end = self._selection.pixels()
                c.rectangle(start, 0, end - start, height)
                c.fill()
            
        context.set_source_surface(self._cache, 0, 0)
        context.set_operator(cairo.OPERATOR_SOURCE)
        context.paint()


class SelectionLayer(object):
    """A layer for LayeredGraphView.

    It highlights the selected area. The cairo surface is cached, it
    is redrawn only when selection has changed.

    """
    def __init__(self, layered, selection):
        self._layered = layered
        self._cache = None
        self._selection = selection
        self._selection.changed.connect(self.update)

    def update(self):
        self._cache = None
        self._layered.redraw()

    def draw(self, context, width, height):
        if not self._cache:
            surface = context.get_target()
            self._cache = surface.create_similar(cairo.CONTENT_COLOR_ALPHA,
                                                 width, height)
            c = cairo.Context(self._cache)

            if self._selection.selected():
                start, end = self._selection.pixels()
                # darken everything ...
                c.set_source_rgba(0, 0, 0, 0.5)
                c.paint()

                # ... then clear selection
                c.set_operator(cairo.OPERATOR_CLEAR)
                c.rectangle(start, 0, end - start, height)
                c.fill()
            
        context.set_source_surface(self._cache, 0, 0)
        context.set_operator(cairo.OPERATOR_OVER)
        context.paint()

class CursorLayer(object):

    def __init__(self, layered, cursor):
        self._layered = layered
        self._cache = None
        self._cursor = cursor
        self._cursor.changed.connect(self.update)
        self._height = None
        self.rgba = (1, 1, 1, 0.5)

    def update(self):
        self._layered.redraw() # called in all layers !

    def draw(self, context, width, height):
        if height != self._height:
            self._height = height
            surface = context.get_target()
            self._cache = surface.create_similar(cairo.CONTENT_COLOR_ALPHA,
                                                 1, height)
            c = cairo.Context(self._cache)
            c.set_source_rgba(*self.rgba)
            c.set_line_width(1)
            c.move_to(0.5, 0)
            c.line_to(0.5, height)
            c.stroke()

        x = self._cursor.pixel()
        context.set_source_surface(self._cache, x, 0)
        context.set_operator(cairo.OPERATOR_OVER)
        context.paint()


# -- Mouse event listeners that act on models.
#

def near(x, y):
    return abs(x - y) < 10

class MouseScroll(object):
    """Listens for mouse wheel events and scroll a graph.

    Must be attached to a gtk.Widget and a Graph.

    """
    def __init__(self, widget, graph):
        self._graph = graph
        widget.add_events(gtk.gdk.SCROLL_MASK)
        widget.connect("scroll_event", self.scroll_event)

    def scroll_event(self, widget, event):
        MOD1 = event.state & gtk.gdk.MOD1_MASK
        LEFT = event.direction is gtk.gdk.SCROLL_LEFT
        RIGHT = event.direction is gtk.gdk.SCROLL_RIGHT
        UP = event.direction is gtk.gdk.SCROLL_UP
        DOWN = event.direction is gtk.gdk.SCROLL_DOWN

        if LEFT or (UP and MOD1):
            self._graph.scroll_left()
        elif RIGHT or (DOWN and MOD1):
            self._graph.scroll_right()
        elif UP:
            self._graph.zoom_in_on(event.x)
        elif DOWN:
            self._graph.zoom_out_on(event.x)


class MouseSelection(object):
    """Listens for mouse events and select graph area.

    Must be attached to a gtk.Widget and a Selection.

    """
    def __init__(self, widget, selection):
        self.widget = widget
        self._selection = selection
        self.pressed = False
        widget.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                          gtk.gdk.BUTTON_RELEASE_MASK |
                          gtk.gdk.POINTER_MOTION_MASK |
                          gtk.gdk.POINTER_MOTION_HINT_MASK)
        widget.connect("button_press_event", self.button_press)
        widget.connect("button_release_event", self.button_release)
        widget.connect("motion_notify_event", self.motion_notify)

    def button_press(self, widget, event):
        if event.button == 1:
            self.pressed = True
            x = event.x
            start, end = self._selection.pixels()
            # a double click resumes selection.
            if event.type == gtk.gdk._2BUTTON_PRESS:
                self._selection.pin(x)
            # extend towards left
            elif self._selection.selected() and near(start, x):
                self._selection.pin(end)
                self._selection.extend(x)
            # extend towards right
            elif self._selection.selected() and near(end, x):
                self._selection.pin(start)
                self._selection.extend(x)
            # start fresh selection
            else:
                self._selection.pin(x)

    def button_release(self, widget, event):
        if event.button == 1:
            self.pressed = False

    def motion_notify(self, widget, event):
        if self.pressed:
            x = event.window.get_pointer()[0]
            self._selection.extend(x)


class MouseMiddleClick(object):
    """Shift the wave display when the middle button is pressed."""
    def __init__(self, widget, graph):
        self.widget = widget
        self.graph = graph
        self.pressed = False
        widget.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                          gtk.gdk.BUTTON_RELEASE_MASK |
                          gtk.gdk.POINTER_MOTION_MASK |
                          gtk.gdk.POINTER_MOTION_HINT_MASK)
        widget.connect("button_press_event", self.button_press)
        widget.connect("button_release_event", self.button_release)
        widget.connect("motion_notify_event", self.motion_notify)

    def button_press(self, widget, event):
        if event.button == 2:
            self.pressed = True
            self._xlast = event.x

    def button_release(self, widget, event):
        if event.button == 2:
            self.pressed = False

    def motion_notify(self, widget, event):
        if self.pressed:
            x = event.window.get_pointer()[0]
            delta = self._xlast - x
            self._xlast = x
            start, _ = self.graph.view()
            self.graph.move_to(start + delta * self.graph.density())


class PointerStyle(object):
    """Change the pointer style.

    Show a hand when the middle click is pressed, otherwise show
    special cursors if the pointer is next to a selection bound.

    """
    LEFT_SIDE = gtk.gdk.Cursor(gtk.gdk.LEFT_SIDE)
    RIGHT_SIDE = gtk.gdk.Cursor(gtk.gdk.RIGHT_SIDE)
    HAND = gtk.gdk.Cursor(gtk.gdk.HAND1)

    def __init__(self, widget, selection):
        self.widget = widget
        self._selection = selection
        self.pressed = False
        widget.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                          gtk.gdk.BUTTON_RELEASE_MASK |
                          gtk.gdk.POINTER_MOTION_MASK |
                          gtk.gdk.POINTER_MOTION_HINT_MASK)
        widget.connect("button_press_event", self.button_press)
        widget.connect("button_release_event", self.button_release)
        widget.connect("motion_notify_event", self.motion_notify)

    def button_press(self, widget, event):
        if event.button == 2:
            self.pressed = True
            self.widget.window.set_cursor(self.HAND)

    def button_release(self, widget, event):
        if event.button == 2:
            self.pressed = False
            self.widget.window.set_cursor(None)

    def motion_notify(self, widget, event):
        if not self.pressed:
            style = None
            x = event.window.get_pointer()[0]
            if self._selection.selected():
                start, end = self._selection.pixels()
                if near(start, x):
                    style = self.LEFT_SIDE
                elif near(end, x):
                    style = self.RIGHT_SIDE
            self.widget.window.set_cursor(style)


# -- Horizontal scrollbar, subclassed to control a Graph object.
#
class GraphScrollbar(gtk.HScrollbar):
    """An horizontal scrollbar tied to a Graph.

    Acts on a graph object and is updated when the graph
    object changes.

    """
    def __init__(self, graph):
        self._adjustment = gtk.Adjustment(0, 0, 1, 0.1, 0, 1)
        super(GraphScrollbar, self).__init__(self._adjustment)
        self._graph = graph
        self._graph.changed.connect(self.update_scrollbar)
        self.connect("value-changed", self.update_model)
        
        # When the scrollbar or the graph model changes, this
        # attribute must be set to True to avoid infinite feedback
        # between them. Example of what happens otherwise: scrollbar
        # changes --> graph changes --> scrollbar changes --> ...
        self.inhibit = False

    def update_model(self, widget):
        """Changes the model.

        Called when the scrollbar has been moved by the user.

        """
        if not self.inhibit:
            self.inhibit = True
            self._graph.move_to(self._adjustment.value)
            self.inhibit = False
        
    def update_scrollbar(self):
        """Changes the scrollbar.

        Called when the model has changed.

        """
        if not self.inhibit:
            self.inhibit = True
            length = self._graph.numframes()
            start, end = self._graph.view()
            if start != end:
                page_size = (end - start)
                self._adjustment.set_all(value = start,
                                         lower = 0,
                                         upper = length,
                                         page_increment = page_size,
                                         step_increment = page_size / 5.,
                                         page_size = page_size)
            else:
                # empty sound
                self._adjustment.set_all(value = 0,
                                         lower = 0,
                                         upper = 1,
                                         page_increment = 0,
                                         step_increment = 0,
                                         page_size = 1)
            self.inhibit = False


# -- Tests

if __name__ == '__main__':
    from mock import Mock, Fake

    def test_layered():

        def randomized():
            from random import random        
            channels = [[((random() - 0.5) * 2, (random() - 0.5) * 2)
                       for i in xrange(500)]]
            graph = Mock({"channels": channels,
                          "set_width": None,
                          "frames_info": (0, 0, 0)})
            graph.changed = Fake()
            layered = LayeredGraphView(graph)
            layered.layers.append(WaveformLayer(layered, graph))
            return layered

        def sine():
            from math import sin
            sine = [sin(2 * 3.14 * 0.01 * x) for x in xrange(500)]
            channels = [[(i, i) for i in sine]]
            graph = Mock({"channels": channels, "set_width": None,
                          "frames_info": (0, 0, 0)})
            graph.changed = Fake()
            layered = LayeredGraphView(graph)
            layered.layers.append(WaveformLayer(layered, graph))
            return layered

        def sines():
            from math import sin
            sine = [sin(2 * 3.14 * 0.01 * x) for x in xrange(500)]
            channels = [[(i, i) for i in sine], [(i, i) for i in sine]]

            graph = Mock({"channels": channels, "set_width": None,
                          "frames_info": (0, 0, 0)})
            graph.changed = Fake()
            layered = LayeredGraphView(graph)
            layered.layers.append(WaveformLayer(layered, graph))
            return layered

        def test_selection_layer(layerclass):
            graph = Mock({"channels": [], "set_width": None,
                          "frames_info": (0, 0, 0)})
            graph.changed = Fake()
            selection = Mock({"pixels": (20, 100), "selected": True})
            selection.changed = Fake()
            layered = LayeredGraphView(graph)
            layered.layers.append(layerclass(layered, selection))
            return layered

        def background():
            return test_selection_layer(BackgroundLayer)

        def selection():
            return test_selection_layer(SelectionLayer)

        def cursor():
            graph = Mock({"channels": [], "set_width": None,
                          "frames_info": (0, 0, 0)})
            class Cursor: pass
            cursor = Cursor()
            cursor.changed = Fake()
            cursor.pixel = lambda: 50
            layered = LayeredGraphView(graph)
            cursorlayer = CursorLayer(layered, cursor)
            cursorlayer.rgba = (1, 0, 0, 1)
            layered.layers.append(cursorlayer)
            return layered

        layereds = [randomized(), sine(), sines(), selection(), cursor(),
                    background()]

        for layered in layereds:
            window = gtk.Window()
            window.resize(500, 200)
            window.connect("delete-event", gtk.main_quit)
            window.add(layered)
            window.show_all()
            gtk.main()

    test_layered()



"""A sound editor.

Scalpel is a sound editor. It works only on Linux at this point, with
a GTK frontend and an ALSA backend.

Changelog for this release at: http://scalpelsound.online.fr/?p=388

"""
from constants import __version__



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

from event import Signal
from threading import Thread, Event, Lock

# Cursor position can be set by a Selection object. When Player is
# playing, Cursor regularly updates itself with its own thread.

class Repeat(Thread):
    """Regularly call a funtion.

    This object is a thread that calls a function repeatedly, with a
    delay between each call.

    """
    def __init__(self, function, delay):
        Thread.__init__(self)
        self._function = function
        self.delay = delay
        self.must_stop = Event()

    def run(self):
        while not self.must_stop.isSet():
            self._function()
            self.must_stop.wait(self.delay)

    def stop(self):
        """ Stop the thread. 

        Does not return until the thread is actually stopped.

        """
        self.must_stop.set()
        self.join()


class Cursor(object):
    """Cursor position (frame and pixel).

    When the player is playing, pixel() returns the player position,
    otherwise the position that was set (through a Selection object)
    is returned.

    """
    def __init__(self, graph, player):
        self._graph = graph
        self._player = player
        # Position set through a Selection object
        self._frame = 0
        # Position of the player
        self._player_frame = 0
        self._pixel = 0
        self.changed = Signal()
        self._player.start_playing.connect(self._on_start_playing)
        self._player.stop_playing.connect(self._on_stop_playing)
        self._graph.changed.connect(self._on_graph_changed)
        self._follow = None
        self._lock = Lock()

    def pixel(self):
        return self._pixel

    def set_frame(self, frame):
        """This method may be called by a Selection object."""
        self._frame = frame
        if not self._follow:
            self._update_pixel(frame)

    def _set_player_frame(self, frame):
        """This method is only called by self._follow thread."""
        self._player_frame = frame
        self._update_pixel(frame)

    # This method may be called concurrently. A lock ensures
    # atomicity.
    def _update_pixel(self, frame):
        self._lock.acquire()
        pixel = self._graph.frmtopxl(frame)
        if pixel != self._pixel:
            self._pixel = pixel
            self.changed()
        self._lock.release()

    def _on_graph_changed(self):
        if self._follow:
            self._update_pixel(self._player_frame)
        else:
            self._update_pixel(self._frame)

    def _on_start_playing(self):
        if not self._follow:
            self._follow = Repeat(self._store_player_position, 0.05)
            self._follow.start()

    def _on_stop_playing(self):
        if self._follow:
            self._follow.stop()
            self._follow = None
        self.set_frame(self._frame)

    def _store_player_position(self):
        self._set_player_frame(self._player.position)


if __name__ == "__main__":

    def test():
        import time
        from mock import Fake

        class Empty: pass

        graph = Empty()
        graph.changed = Fake()
        graph.frmtopxl = lambda x: x
        player = Empty()
        player.start_playing = Fake()
        player.stop_playing = Fake()
        c = Cursor(graph, player)

        c.set_frame(5)
        assert c.pixel() == 5

        player.position = 0
        c._on_start_playing()
        time.sleep(0.2)
        assert c.pixel() == 0
        c.set_frame(10)
        assert c.pixel() == 0
        c._on_stop_playing()
        time.sleep(0.2)
        assert c.pixel() == 10

    test()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

from event import Signal

def _overview(data, density):
    "Returns a list of (min, max) tuples."
    if density < 1:
        density = 1
    left = 0
    end = len(data)
    res = []
    while round(left) < end:
        right = left + density
        if right > end:
            right = end
        i = int(round(left))
        j = int(round(right))
        d = data[i:j]
        mini = d.min()
        maxi = d.max()
        res.append((mini, maxi))
        left = right
    return res

class Graph(object):
    """Scale the sound visualization.
    
    When audio is drawn on the screen, several frames are condensed in
    one column of pixels. This object computes what to display,
    according to zooming and position in the sound.

    """
    def __init__(self, sound):
        self.changed = Signal()
        self._width = 100
        self.set_sound(sound)

    def set_sound(self, sound):
        self._sound = sound
        self._view_start = 0
        self._view_end = len(self._sound.frames)
        self._sound.changed.connect(self.update)
        self.changed()

    def update(self):
        self._adjust_view()
        self.changed()

    def set_view(self, start, end):
        self._view_start = start
        self._view_end = end
        self.update()

    def move_to(self, frame):
        "Moves the view start and keep the view length"
        l = self._view_end - self._view_start
        self.set_view(frame, frame + l)

    def center_on(self, frame):
        l = self._view_end - self._view_start
        self.set_view(frame - l * 0.5, frame + l * 0.5)

    def numframes(self):
        return len(self._sound.frames)
    
    def view(self):
        start = self._view_start
        end = self._view_end
        return (start, end)

    def density(self):
        "Number of frames per pixel."
        number_frames_view = (self._view_end - self._view_start)
        d = float(number_frames_view) / self._width
        if d < 1:
            d = 1
        return d

    def frmtopxl(self, f):
        "Converts a frame index to a pixel index."
        p = int(round(f - self._view_start) / self.density())
        return p

    def pxltofrm(self, p):
        "Converts a pixel index to a frame index."
        f = int(round(self._view_start + p * self.density()))
        return self._gauge(f, 0, self.numframes())
    
    def _gauge(self, value, mini, maxi):
        "Calibrate value between mini and maxi."
        if value < mini:
            value = mini
        if value > maxi:
            value = maxi
        return value

    def _zoom(self, factor):
        """Expand or shrink view according to factor.

        0 < factor < 1     zoom in
            factor = 1     unchanged
        1 < factor < +Inf  zoom out

        The zoom factor is relative to the current zoom, ie.::

            self._zoom(x, n)
            self._zoom(x, m)

        is equivalent to::

            self._zoom(x, n * m)
        
        """
        l = self._view_end - self._view_start
        l = l * factor
        self._view_end = self._view_start + l
        
    def zoom_in(self):
        "Make view twice smaller, centering on the middle of the view."
        mid = self._view_start + (self._view_end - self._view_start) * 0.5
        self._zoom(0.5)
        self.center_on(mid)

    def zoom_out(self):
        "Make view twice larger, centering on the middle of the view."
        mid = self._view_start + (self._view_end - self._view_start) * 0.5
        self._zoom(2)
        self.center_on(mid)

    def zoom_out_full(self):
        "Fit everything in the view."
        self.set_view(0, len(self._sound.frames))

    def is_zoomed_out_full(self):
        start, end = self.view()
        return start == 0 and end == len(self._sound.frames)

    def zoom_in_on(self, pixel):
        point = self.pxltofrm(pixel)
        self.zoom_in()
        self.move_to(point - pixel * self.density())

    def zoom_out_on(self, pixel):
        point = self.pxltofrm(pixel)
        self.zoom_out()
        self.move_to(point - pixel * self.density())

    def _scroll(self, factor):
        """Shift the view.

        A negative factor shift the view to the left, a positive one
        to the right. The absolute value of the factor determines the
        length of the shift, relative to the view length. For example:
        0.1 is 10%, O.5 is one half, 1.0 is 100%.
        
        """
        l = self._view_end - self._view_start
        l = int(l * factor)
        self.set_view(self._view_start + l, self._view_end + l)

    def scroll_left(self):
        self._scroll(-0.20)
            
    def scroll_right(self):
        self._scroll(0.20)

    def set_width(self, width):
        "Set the number of values the graph must have."
        self._width = int(width)
        self.update()
        
    def channels(self):
        "Return the graph values."
        width = self._width
        numchan = self._sound.numchan()
        start, end = [int(round(v)) for v in self._view_start, self._view_end]
        visible = self._sound.frames[start:end]
        if numchan == 1:
            o = [_overview(visible, self.density())]
        else:
            visible = visible.transpose()
            o = []
            for chan in range(numchan):
                values = _overview(visible[chan], self.density())
                o.append(values)
        return o

    def _adjust_view(self):
        """Adjust view boundaries.

        After scrolling or zooming, the view must be adjusted. Indeed,
        it must not start before 0 or end after the sound. Also, the
        view must contain enough values to fit in the width.

        """
        width = self._width

        # When the zoom is too strong and the width cannot be filled,
        # zoom out.
        view_length = self._view_end - self._view_start
        if view_length < width:
            self._view_start -= (width - view_length) * 0.5
            self._view_end = self._view_start + width

        # Check and adjust visible bounds.
        if self._view_start < 0:
            self._view_end += -self._view_start 
            self._view_start = 0
        elif self._view_end > len(self._sound.frames):
            self._view_start -= (self._view_end - len(self._sound.frames))
            self._view_end = len(self._sound.frames)

        # Ultimate check on bounds.
        if self._view_start < 0:
            self._view_start = 0
        if self._view_end > len(self._sound.frames):
            self._view_end = len(self._sound.frames)

       
def test_overview():
    import numpy
    b = numpy.array(range(1000000))
    assert len(_overview(b, 100000)) == 10
    assert len(_overview(b, 10000)) == 100
    
def test_Graph():
    from mock import Mock, Fake
    import numpy
    
    sound = Mock({"numchan": 1})
    sound.changed = Fake()
    sound.frames = numpy.array(range(1000))

    c = Graph(sound)
    c.set_width(200)
    o = c.channels()

    class Foo:
        def foo(self):
            print "Changed."
    f = Foo()
    c = Graph(sound)
    c.changed.connect(f.foo)

    c.set_width(200)
    o = c.channels()

    # stereo
    import numpy
    sound = Mock({"numchan": 2})
    sound.changed = Fake()
    data = numpy.array([[1, 1], [2, 2], [3, 3]])
    sound.frames = data
    c = Graph(sound)
    o = c.channels()
    assert(len(o)) == 2
    

def test_zoom():
    from mock import Mock, Fake
    import numpy

    sound = Mock({"numchan": 1})
    data = numpy.array([1, 2, 3, 4])
    sound.frames = data
    sound.changed = Fake()

    g = Graph(sound)

    g.set_width(4)
    g._zoom(1)
    g.center_on(1.5)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]
    
    g._zoom(factor=1)
    g.center_on(0)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]

    g._zoom(1)
    g.center_on(6)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]
    
    g.set_width(2)
    g._zoom(0.5)
    g.center_on(1.5)
    o = g.channels()
    assert o == [[(2, 2), (3, 3)]]

    g._zoom(factor=0.5)
    g.center_on(1.5)
    g.set_width(4)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]

    g.set_width(2)
    g._zoom(0.5)
    g.center_on(0)
    o = g.channels()
    assert o == [[(1, 1), (2, 2)]]

    g.set_width(4)
    g._zoom(0.25)
    g.center_on(0)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]
    
    g.set_width(4)
    g._zoom(4)
    g.center_on(4)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]

    g.set_width(3)
    data = numpy.array([1, 2, 3, 4, 5])
    sound.frames = data
    g._zoom(0.5)
    g.center_on(2)
    o = g.channels()
    assert o == [[(2, 2), (3, 3), (4, 4)]]

    g._zoom(factor=0.5)
    g.center_on(2)
    g._zoom(factor=0.5)
    g.center_on(2)
    start, end = g._view_start, g._view_end
    
    g._zoom(factor=0.5 * 0.5)
    g.center_on(2)
    assert (start, end) == (g._view_start, g._view_end)

def test_zoom_in():
    import numpy 
    from mock import Mock, Fake
    sound = Mock({"numchan": 1})
    sound.changed = Fake()

    data = numpy.array([1, 2, 3, 4])
    sound.frames = data
    g = Graph(sound)

    g.set_width(2)
    g.zoom_in()
    o = g.channels()
    assert o == [[(2, 2), (3, 3)]]

    g.zoom_out()
    g.set_width(4)
    o = g.channels()
    assert o == [[(1, 1), (2, 2), (3, 3), (4, 4)]]

def test_zoom_in_on():
    import numpy
    from mock import Mock, Fake
    sound = Mock({"numchan": 1})
    sound.changed = Fake()
    data = numpy.array([1, 2, 3, 4])
    sound.frames = data
    g = Graph(sound)
    g.set_width(2)

    g.zoom_in_on(0)
    assert g.channels() == [[(1, 1), (2, 2)]]

    g.zoom_out()
    g.zoom_in_on(1)
    assert g.channels() == [[(2, 2), (3, 3)]]

    g.zoom_out()
    g.zoom_in_on(2)
    assert g.channels() == [[(3, 3), (4, 4)]]

def test_scroll():
    import numpy
    from mock import Mock, Fake

    sound = Mock({})
    data = numpy.array([1, 2, 3, 4])
    sound.frames = data
    sound.changed = Fake()

    g = Graph(sound)
    g.set_width(4)

    g.scroll_right()
    length = g.numframes()
    start, end = g.view()
    assert length == 4
    assert start == 0
    assert end == 4

    
if __name__ == "__main__":
    test_overview()
    test_Graph()
    test_zoom()
    test_zoom_in()
    test_scroll()
    test_zoom_in_on()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import edit
import clipboard
from player import Player
from event import Signal
import effect
import app
import traceback

class Controller(object):

    def __init__(self, sound, player, graph, selection):
        self._player = player
        self._graph = graph
        self._selection = selection
        self._sound = sound
        self.filename_changed = Signal()
        self.error = Signal()
        self.clipboard = clipboard.Clipboard()

    def new(self):
        app.open_()

    def _report_exception(method):
        """Method decorator.

        Call the self.error signal when an exception was raised in the
        decorated method.

        """
        def wrapper(self, *args, **kwargs):
            try:
                method(self, *args, **kwargs)
            except Exception, e:
                self.error("Error", str(e))
                traceback.print_exc()

        wrapper.__name__ = method.__name__
        wrapper.__doc__ = method.__doc__
        return wrapper

    @_report_exception
    def open(self, filename):
        if self._sound.is_fresh():
            self.load_sound(filename)
        else:
            app.open_(filename)

    @_report_exception
    def load_sound(self, filename):
        self._player.pause()
        self._sound = edit.Sound(filename)
        self._graph.set_sound(self._sound)
        self._player.set_sound(self._sound)
        self._selection.unselect()
        self.filename_changed()

    @_report_exception
    def save(self):
        self._sound.save()

    @_report_exception
    def save_as(self, filename):
        self._sound.save_as(filename)
        self.filename_changed()

    @_report_exception
    def save_selection_as(self, filename):
        sound = edit.Sound()
        if not self._selection.selected():
            raise Exception("There is no selection.")
        start, end = self._selection.get()
        sound.frames = self._sound.copy(start, end)
        sound.samplerate = self._sound.samplerate
        sound.save_as(filename)

    def close(self, force=False):
        sound = self._sound
        if not sound.is_saved() and not sound.is_fresh() and not force:
            raise FileNotSaved
        self._player.pause()

    def play(self):
        start, end = self._selection.get()
        if not self._selection.selected():
            end = len(self._sound.frames)
        self._player.start = start
        self._player.end = end
        self._player.thread_play()

    def pause(self):
        self._player.pause()

    def goto_start(self):
        self._selection.set(0, 0)

    def goto_end(self):
        end = len(self._sound.frames)
        self._selection.set(end, end)

    def cut(self):
        start, end = self._selection.get()
        self._selection.start = start
        self._selection.end = start
        self.clipboard.clip = self._sound.cut(start, end)
        
    def copy(self):
        start, end = self._selection.get()
        self.clipboard.clip = self._sound.copy(start, end)
        
    @_report_exception
    def paste(self):
        start, end = self._selection.get()
        was_zoomed_out_full = self._graph.is_zoomed_out_full()
        self._sound.paste(start, end, self.clipboard.clip)
        self._selection.set(start, start + len(self.clipboard.clip))
        if was_zoomed_out_full:
            self._graph.zoom_out_full()

    @_report_exception
    def mix(self):
        start, end = self._selection.get()
        self._sound.mix(start, end, self.clipboard.clip)
        # FIXME: should be dealt with in Sound.
        if start == end:
            l = len(self.clipboard.clip)
        else:
            l = min(end - start, len(self.clipboard.clip))
        self._selection.set(start, start + l)

    def trim(self):
        pass

    def undo(self):
        self._sound.undo()

    def redo(self):
        self._sound.redo()

    def select_all(self):
        self._selection.select_all()

    def zoom_out(self):
        self._graph.zoom_out()

    def zoom_in(self):
        self._graph.zoom_in()

    def zoom_fit(self):
        if self._selection.selected():
            start, end = self._selection.get()
            self._graph.set_view(start, end)
        else:
            self._graph.zoom_out_full()

    def list_effects(self):
        l = effect.effects.keys()
        l.sort()
        return l

    def list_extensions(self):
        return edit.list_extensions()

    def effect(self, name):
        if self._selection.selected():
            start, end = self._selection.get()
        else:
            start = 0
            end = len(self._sound.frames)
        fx = effect.effects[name]
        fx(self._sound, start, end)

    def filename(self):
        return self._sound.filename

class FileNotSaved: pass

def test_Controller():
    from mock import Fake
    
    # Test opening a file
    ctrl = Controller(Fake(), Fake(), Fake(), Fake())
    ctrl.open('../sounds/test1.wav')
    assert ctrl._sound != None

if __name__ == "__main__":
    test_Controller()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import constants

# pygtk gets program name from sys.argv[0]. This name appears in
# taskbars when windows are grouped together.
import sys
sys.argv[0] = constants.__appname__

import app
import control
from gtkwaveform import GraphView, GraphScrollbar
from gtkfiledialog import FileDialog
import copy
import os.path
import urllib
import gtk
gtk.gdk.threads_init()

def init():
    """Called when the module is being imported."""
    # Plug callbacks into app.
    app.new_sound_loaded.connect(on_new_sound_loaded)

def main_loop():
    gtk.main()

def on_new_sound_loaded(controller, graph, sel, curs):
    EditorWindow(controller, graph, sel, curs)

def display_error(title, text, parent=None):
    d = gtk.MessageDialog(parent, buttons=gtk.BUTTONS_CLOSE)
    d.set_title(title)
    d.set_markup(text)
    d.run()
    d.destroy()


class EditorWindow(gtk.Window):

    _windows = []

    def __init__(self, controller, graph, selection, cursor):
        gtk.Window.__init__(self)

        self.ctrl = controller
        
        self.uimanager = self._make_ui_manager()
        accelgroup = self.uimanager.get_accel_group()
        self.add_accel_group(accelgroup)

        self.menubar = self.uimanager.get_widget('/menubar')
        self.toolbar = self.uimanager.get_widget('/toolbar')
        self.toolbar.set_style(gtk.TOOLBAR_ICONS)
        for w in self.toolbar:
            w.set_homogeneous(False)
            w.set_focus_chain([])

        # Create and fill the Effects menu
        effect_menu = gtk.Menu()
        for name in controller.list_effects():
            item = gtk.MenuItem(label=name)
            item.connect('activate', self.effect, None, name)
            effect_menu.append(item)
        w = self.uimanager.get_widget('/menubar/Effects')
        w.set_submenu(effect_menu)

        self.waveform = GraphView(graph, selection, cursor)
        self.scrollbar = GraphScrollbar(graph)
        self.statusbar = gtk.Statusbar()
        self.filedialog = FileDialog(controller.list_extensions())

        self.vbox = gtk.VBox()
        self.vbox.pack_start(self.menubar, expand=False, fill=False)
        self.vbox.pack_start(self.toolbar, expand=False, fill=False)
        self.vbox.pack_start(self.waveform, expand=True, fill=True)
        self.vbox.pack_start(self.scrollbar, expand=False, fill=False)
        self.vbox.pack_end(self.statusbar, expand=False, fill=False)
        self.add(self.vbox)
        
        # Setup drag and drop
        TARGET_TYPE_TEXT = 80
        self.drag_dest_set(gtk.DEST_DEFAULT_ALL,
                           [("text/uri-list", 0, TARGET_TYPE_TEXT)],
                           gtk.gdk.ACTION_COPY)
        self.connect("drag_data_received", self._open_dropped_files)

        self.connect("delete-event", self.close)
        self._filename_update()
        self.ctrl.filename_changed.connect(self._filename_update)
        self.ctrl.error.connect(self.display_error)
        icon = self.render_icon(gtk.STOCK_CUT, gtk.ICON_SIZE_MENU)
        self.set_icon(icon)
        self.resize(700, 500)
        self.show_all()
        self._windows.append(self)

    def _make_ui_manager(self):
        ui = '''<ui>
            <menubar name="menubar">
              <menu action="File">
                <menuitem action="New"/>
                <menuitem action="Open"/>
                <menuitem action="Save"/>
                <menuitem action="Save as"/>
                <menuitem action="Save Selection as"/>
                <separator/>
                <menuitem action="Close"/>
                <menuitem action="Quit"/>
              </menu>
              <menu action="Edit">
                <menuitem action="Undo"/>
                <menuitem action="Redo"/>
                <separator/>
                <menuitem action="Cut"/>
                <menuitem action="Copy"/>
                <menuitem action="Paste"/>
                <menuitem action="Mix"/>
                <separator/>
                <menuitem action="SelectAll"/>
              </menu>
              <menu action="View">
                <menuitem action="ZoomIn"/>
                <menuitem action="ZoomOut"/>
                <menuitem action="ZoomFit"/>
              </menu>
              <menu action="Effects">
              </menu>                
              <menu action="Scalpel">
                <menuitem action="About"/>
              </menu>
            </menubar>
            <toolbar name="toolbar">
              <toolitem action="Open"/>
              <separator/>
              <toolitem action="ZoomFit"/>
              <toolitem action="ZoomOut"/>
              <toolitem action="ZoomIn"/>
              <separator/>
              <toolitem action="Start"/>
              <toolitem action="Pause"/>
              <toolitem action="Play"/>
              <toolitem action="End"/>
              <separator/>
              <toolitem action="Cut"/>
              <toolitem action="Copy"/>
              <toolitem action="Paste"/>
              <toolitem action="Mix"/>
              <separator/>
              <toolitem action="Undo"/>
              <toolitem action="Redo"/>
            </toolbar>
            <accelerator action="Start"/>
            <accelerator action="End"/>
            <accelerator action="Play"/>
            <accelerator action="Pause"/>
        </ui>'''

        uimanager = gtk.UIManager()
        uimanager.add_ui_from_string(ui)

        # An action is a tuple formed like this:
        # (action_name, stock_id, label, accelerator, tooltip, callback)
        #
        # Actions that are not in a menu (e.g. 'Play') must have an
        # <accelerator> tag in the UI description for the accelerator
        # defined below to be effective.

        actions = [('File', None, '_File'),
                   ('Edit', None, '_Edit'),
                   ('View', None, '_View'),
                   ('Effects', None, '_Effects'),
                   ('Scalpel', None, '_Scalpel'),
                   ('New', gtk.STOCK_NEW, None, None, '', self.new),
                   ('Open', gtk.STOCK_OPEN, None, None, '', self.open),
                   ('Save', gtk.STOCK_SAVE, None, None, '', self.save),
                   ('Save as', gtk.STOCK_SAVE_AS, None, None, '',self.save_as),
                   ('Save Selection as', gtk.STOCK_SAVE_AS,'Save Selection As',
                                 '<Ctrl><Alt>s', None, self.save_selection_as),
                   ('Close', gtk.STOCK_CLOSE, None, None, '', self.close),
                   ('Quit', gtk.STOCK_QUIT, None, None, '', self.quit),
                   ('Play', gtk.STOCK_MEDIA_PLAY, None, 'p', '', self.play),
                   ('Pause', gtk.STOCK_MEDIA_PAUSE, None, 'o', '', self.pause),
                   ('Start', gtk.STOCK_MEDIA_PREVIOUS, None, 'Home', '',
                                                              self.goto_start),
                   ('End', gtk.STOCK_MEDIA_NEXT, None, 'End', '',
                                                                self.goto_end),
                   ('Cut', gtk.STOCK_CUT, None, None, '', self.cut),
                   ('Copy', gtk.STOCK_COPY, None, None, '', self.copy),
                   ('Paste', gtk.STOCK_PASTE, None, None, '', self.paste),
                   ('Mix', gtk.STOCK_ADD, 'Mix', '<Ctrl><Shift>v', None,
                                                                     self.mix),
                   ('Undo', gtk.STOCK_UNDO, None, '<Ctrl>z', None, self.undo),
                   ('Redo', gtk.STOCK_REDO, None, '<Ctrl><Shift>z', None,
                                                                    self.redo),
                   ('SelectAll', gtk.STOCK_SELECT_ALL, None, '<Ctrl>a', '',
                                                         self.select_all),
                   ('ZoomFit', gtk.STOCK_ZOOM_FIT, None, 'equal', '',
                                                                self.zoom_fit),
                   ('ZoomOut', gtk.STOCK_ZOOM_OUT, None, 'KP_Subtract', '',
                                                                self.zoom_out),
                   ('ZoomIn', gtk.STOCK_ZOOM_IN, None, 'KP_Add', '',
                                                                 self.zoom_in),
                   ('About', gtk.STOCK_ABOUT, None, None, '', self.about)
                   ]
        actiongroup = gtk.ActionGroup('')
        actiongroup.add_actions(actions)
        uimanager.insert_action_group(actiongroup, 0)

        return uimanager

    def _close(self):
        try:
            self.ctrl.close()
        except control.FileNotSaved:
            proceed = self._show_dialog_close()
            if not proceed:
                return False
            self.ctrl.close(force=True)

        self.destroy()
        self._windows.remove(self)

        if not self._windows:
            self.quit()

        return True

    def _show_dialog_close(self):
        dialog = gtk.MessageDialog(parent=self, type=gtk.MESSAGE_WARNING)
        name = self.ctrl.filename() or "sound"
        name = os.path.basename(name)
        dialog.set_markup("<b>Save %s before closing?</b>" % name)
        dialog.add_button("Close _without saving", gtk.RESPONSE_NO)
        dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        dialog.add_button(gtk.STOCK_SAVE_AS, gtk.RESPONSE_YES)
        response = dialog.run()
        dialog.destroy()
        proceed = False
        if response == gtk.RESPONSE_NO:
            proceed = True
        elif response == gtk.RESPONSE_YES:
            saved = self.save_as()
            if saved:
                proceed = True
        return proceed

    def _filename_update(self):
        filename = self.ctrl.filename()
        self._update_title(filename)
        self.filedialog.filename = filename

    def _update_title(self, filename=None):
        title = constants.__appname__
        if filename:
            title = os.path.basename(filename) + ' - ' + title
        self.set_title(title)

    def _open_dropped_files(self, widget, context, x, y,
                            selection, targetType, time):
        """Open files dragged and dropped on the window."""
        filenames = selection.data.split()
        for filename in filenames:
            # Extract the actual filename
            prefix = 'file://'
            if filename.startswith(prefix):
                filename = filename[len(prefix):]
            filename = urllib.unquote(filename)
            self.ctrl.open(filename)

    def display_error(self, title, text):
        display_error(title, text, parent=self)

    # -- GTK Callbacks

    def __getattr__(self, name):
        """Redirect callbacks to the controller.

        The gtk widget passed to the callback (first argument) will
        not be passed to the controller method.
        
        """
        if name in ["new", "save", "play", "pause",
                    "goto_start", "goto_end", "select_all",
                    "cut", "copy", "paste", "mix", "undo", "redo",
                    "zoom_in", "zoom_out", "zoom_fit",
                    "effect"]:
            method = getattr(self.ctrl, name)
            def forward(self, *args):
                method(*args[1:])
            return forward
        else:
            raise AttributeError(name)


    def about(self, *args):
        d = gtk.AboutDialog()
        d.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        d.set_transient_for(self)
        d.set_program_name(constants.__appname__)
        d.set_version(constants.__version__)
        d.set_website(constants.__url__)
        d.set_copyright("(c) Pierre Duquesne <stackp@online.fr>")
        d.set_comments("A sound editor")
        d.run()
        d.destroy()
        
    def open(self, *args):
        filename = self.filedialog.get_filename(action='open')
        if filename != None:
            self.ctrl.open(filename)

    def save_as(self, *args):
        filename = self.filedialog.get_filename(action='save')
        saved = False
        if filename != None:
            self.ctrl.save_as(filename)
            saved = True
        return saved

    def save_selection_as(self, *args):
        # FIXME: title
        filename = self.filedialog.get_filename(action='save')
        if filename != None:
            self.ctrl.save_selection_as(filename)

    def close(self, *args):
        self._close()
        # Tell GTK not to hide the window, we take care of that:
        return True

    def quit(self, *args):
        for win in copy.copy(self._windows):
            if not win._close():
                return
        gtk.main_quit()


# -- Tests
           
def test():
    from mock import Fake, Mock
    graph = Mock({"frames_info":(0, 0, [], []),
                  "channels": [[(0, 0.5)]],
                  "set_width": None,
                  "scroll_left": None,
                  "scroll_right": None})
    graph.changed = Fake()
    selection = Mock({"pixels": [50, 100],
                      "start_selection": None,
                      "end_selection": None,
                      "selected": True})
    selection.changed = Fake()
    cursor = Mock({'pixel': 20})
    cursor.changed = Fake()
    class FakeController(Fake):
        def __init__(self):
            self.filename_changed = Fake()
            self.error = Fake()
    win = EditorWindow(FakeController(), graph, selection, cursor)
    win.resize(700, 500)
    win.show_all()
    gtk.main()

    display_error("Title", "Text")

if __name__ == '__main__':
    test()
else:
    # Module is being imported.
    init()



import numpy

class Clipboard(object):
    # Borg pattern: http://code.activestate.com/recipes/66531/
    __shared_state = {"clip": numpy.array([])}
    def __init__(self):
        self.__dict__ = self.__shared_state




import types

class Mock(object):

    def __init__(self, dico):
        for name, result in dico.items():
            m = self._make_method(name, result)
            setattr(self, name, m)
            
    def _make_method(self, name, result):
        def func(self, *args, **kwargs):
            return result
        
        method = types.MethodType(func, name)
        return method


class Fake(object):

    def _method(self, *args, **kwargs):
        return []

    def __getattr__(self, a):
        return self._method


if __name__ == "__main__":
    m = Mock({"one": 1, "zero": 0})
    assert m.one() == 1
    assert m.one("bla") == 1
    assert m.zero() == 0
    try:
        m.foo()
        assert False
    except AttributeError:
        assert True
        
    f = Fake()
    f.foo()
    f.bar()

    class NotSoFake(Fake):
        def hello(self):
            return "hello"

    n = NotSoFake()
    assert n.hello() == "hello"
    assert n.what() == []



import gtk

class FileDialog(object):
    """Handle a pair of file dialogs (open and save)."""

    def __init__(self, extensions=[]):
        self.filename = None
        self.extensions = extensions

    def get_filename(self, action='open'):
        """Run a dialog and return a filename or None.

        Valid actions are 'open' and 'save'.

        """
        # I used to create the dialogs only once (on object
        # initialization), and hide and show them, but I can not
        # manage to pre-select a filename after a dialog have been
        # used once. I guess file chooser dialogs are just throwaway
        # objects. Thus, recreate them every time.
        if action == 'open':
            chooser = gtk.FileChooserDialog(
                                action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                         gtk.STOCK_OPEN, gtk.RESPONSE_OK))
            title = 'Open file:'
            stock = gtk.STOCK_OPEN

        elif action == 'save':
            chooser = gtk.FileChooserDialog(
                                action=gtk.FILE_CHOOSER_ACTION_SAVE,
                                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                         gtk.STOCK_SAVE, gtk.RESPONSE_OK))
            title = 'Save as:'
            stock = gtk.STOCK_SAVE

        else:
            raise Exception("action must be 'open' or 'save' (got '%s')"
                            % action)

        # Supported files filter
        filter = gtk.FileFilter()
        filter.set_name("Supported files")
        for ext in self.extensions:
            filter.add_pattern("*." + ext)
        chooser.add_filter(filter)
        chooser.set_filter(filter)

        # All files filter
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        chooser.add_filter(filter)

        chooser.set_title(title)
        icon = chooser.render_icon(stock, gtk.ICON_SIZE_MENU)
        chooser.set_icon(icon)

        if self.filename:
            chooser.select_filename(self.filename)
        response = chooser.run()
        filename = chooser.get_filename()
        chooser.destroy()

        # By default, the GTK loop waits until the process is idle to
        # process events. Now, it is very probable that file I/O will
        # be performed right after this method was called and that
        # would delay hiding the dialog until I/O are done. So,
        # process pending events to hide the dialog right now.
        while gtk.events_pending():
            gtk.main_iteration(False)

        if response == gtk.RESPONSE_OK:
            return filename
        else:
            return None


def test():
    d = FileDialog()
    print d.get_filename(action='open')
    print d.get_filename(action='save')
    try:
        d.get_filename(action='nonsense')
    except Exception, e:
        print e
    else:
        assert False

if __name__ == '__main__':
    test()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

from event import Signal
from copy import copy
import pysndfile
import numpy

def list_extensions():
    extensions = pysndfile.supported_format()
    extensions.append('aif')
    return extensions


class Action(object):
    """Describes an action, and a way to revert that action"""
     
    def __init__(self, do, undo):
        """do and undo are tuples in the form (fun, args) where fun is a
        function and args contains the arguments"""
        self._do = do
        self._undo = undo

    def do(self):
        fun, args = self._do
        return fun(*args)

    def undo(self):
        fun, args = self._undo
        return fun(*args)


class History(object):
    "A list of actions, that can be undone and redone."

    def __init__(self):
        self._actions = []
        self._last = -1
        self._counter = 0

    def _push(self, action):
        if self._last < len(self._actions) - 1:
            # erase previously undone actions
            del self._actions[self._last + 1:]
        self._actions.append(action)
        self._last = self._last + 1
        self._counter += 1
        action.number = self._counter
        
    def undo(self):
        if self._last < 0:
            return None
        else:
            action = self._actions[self._last]
            self._last = self._last - 1
            return action.undo()
            
    def redo(self):
        if self._last == len(self._actions) - 1:
            return None
        else:
            self._last = self._last + 1
            action = self._actions[self._last]
            return action.do()

    def add(self, do, undo):
        "Does an action and adds it to history."
        action = Action(do, undo)
        self._push(action)
        return action.do()

    def revision(self):
        if self._last < 0:
            return 0
        else:
            action = self._actions[self._last]
            return action.number


class Sound(object):

    # frames is a numpy.ndarray, as returned by pysndfile
    
    def __init__(self, filename=None):
        self.filename = filename
        self.history = History()
        self.changed = Signal()
        if filename == None:
            # empty sound
            self.frames = numpy.array([])
            self.samplerate = 44100
            self._saved_revision = None
        else:
            f = pysndfile.sndfile(filename)
            nframes = f.get_nframes()
            self.frames = f.read_frames(nframes)
            self.samplerate = f.get_samplerate()
            f.close()
            self._saved_revision = self.history.revision()

    def numchan(self):
        return self.frames.ndim

    def save(self, format=pysndfile.formatinfo()):
        self.save_as(self.filename, format)

    def save_as(self, filename, format=pysndfile.formatinfo()):
        if filename is None:
            raise Exception("No filename")
        f = pysndfile.sndfile(filename, mode='write',
                              format=format,
                              channels=self.numchan(),
                              samplerate=self.samplerate)
        f.write_frames(self.frames)
        f.close()
        self.filename = filename
        self._saved_revision = self.history.revision()

    def cut(self, start, end):
        clip = copy(self.frames[start:end])
        do = (self._do_cut, (start, end))
        undo = (self._do_paste, (start, start, copy(self.frames[start:end])))
        self.history.add(do, undo)
        self.changed()
        return clip
    
    def _do_cut(self, start, end):
        data = numpy.concatenate((self.frames[:start], self.frames[end:]))
        self.frames = data

    def copy(self, start, end):
        clip = copy(self.frames[start:end])
        return clip

    def paste(self, start, end, clip):
        saved = copy(self.frames[start:end])
        clip = mix_channels_auto(clip, self.numchan())
        do = (self._do_paste, (start, end, clip))
        undo = (self._do_paste, (start, start + len(clip), saved))
        self.history.add(do, undo)
        self.changed()

    def _do_paste(self, start, end, clip):
        x = self.frames
        y = numpy.concatenate((x[:start], clip, x[end:]))
        self.frames = y

    def mix(self, start, end, clip):
        saved = copy(self.frames[start:start + len(clip)])
        clip = mix_channels_auto(clip, self.numchan())
        do = (self._do_mix, (start, end, clip))
        undo = (self._do_paste, (start, start + len(clip), saved))
        self.history.add(do, undo)
        self.changed()

    def _do_mix(self, start, end, clip):
        if start != end:
            length = min(end - start, len(clip))
            self.frames[start:start + length] += clip[:length]
        else:
            a = self.frames
            b = clip
            sound_length = max(len(a), start + len(b))
            if self.numchan() > 1:
                c = numpy.zeros((sound_length, self.numchan()))
            else:
                c = numpy.zeros(sound_length)
            c[:len(a)] = a
            c[start:start + len(b)] += b
            self.frames = c

    def undo(self):
        self.history.undo()
        self.changed()

    def redo(self):
        self.history.redo()
        self.changed()

    def is_fresh(self):
        """True if sound is empty and has never been edited."""
        return not len(self.frames) and not self.history._actions

    def is_saved(self):
        return self._saved_revision == self.history.revision()


def mix_channels(frames, gain_lists):
    """Mix channels into a possibly different number of channels.

    * len(gain_lists) defines the number of output channels,
    * gain_lists[n] contains gains to mix input channels into output
      channel number n. Consequently, len(gain_lists[n]) == frames.ndim.
    
    Returns the mixed signal.

    Examples::

          # mono to stereo
          mix_channels(frames, [[1], [1]])

          # stereo to mono
          mix_channels(frames, [[0.5, 0.5]])

    """
    numchan = frames.ndim
    assert len(gain_lists[0]) == numchan

    # special case for monophonic sounds
    if numchan == 1:
        channels = [frames]
    else:
        channels = frames.transpose()

    # apply gains
    out = []
    for gain_list in gain_lists:
        new_channel = numpy.zeros(len(frames))
        for g, channel in zip(gain_list, channels):
            new_channel += g * channel
        out.append(new_channel)

    # special case for monophonic sounds
    if len(out) == 1:
        out = out[0]
        out = numpy.array(out)
    else:
        out = numpy.array(out)
        out = out.transpose()

    return out

def mix_channels_auto(frames, n):
    """Convert the number of channels.

    Used in Sound.paste() and Sound.mix() to convert the clipboard to
    the right number of channels.
    
    This should probably be a Sound method and the clipboard should
    contain a Sound instance instead of just a numpy array.

    Examples::

          # to stereo
          mix_channels_auto(frames, 2)

          # to mono
          mix_channels(frames, 1)

    """
    numchan = frames.ndim
    if n == numchan:
        out = frames
    else:
        if n == 1:
            # monoize
            g = 1. / numchan
            gain_list = [g] * numchan
            gain_lists = [gain_list]
        elif numchan == 1:
            # mono to n channels
            gain_lists = [[1]] * n
        else:
            raise Exception("Don't know yet how to deal with so many channels")
        out = mix_channels(frames, gain_lists)

    return out

# -- Tests

def testAction():
    f = lambda x: x
    do = (f, (1,))
    undo = (f, (2,))
    a = Action(do, undo)
    assert a.do() == 1
    assert a.undo() == 2

    g = lambda : 1
    do = (g, ())
    undo = (g, ())
    a = Action(do, undo)
    assert a.do() == 1
    assert a.undo() == 1

def testHistory():    
    history = History()
    f = lambda x: x
    do = (f, (1,))
    undo = (f, (2,))
    assert history.add(do, undo) == 1
    assert history.undo() == 2
    assert history.undo() == None
    assert history.redo() == 1
    assert history.redo() == None
    assert history.undo() == 2
    assert history.add(do, undo) ==  1
    assert history.add(do, undo) == 1
    assert history.undo() == 2
    assert history.undo() == 2
    assert history.redo() == 1
    assert history.redo() == 1
    assert history.redo() == None

    # Test Revisions
    history = History()
    assert history.revision() == 0
    history.add(do, undo)
    assert history.revision() == 1
    history.add(do, undo)
    assert history.revision() == 2
    history.undo()
    assert history.revision() == 1
    history.redo()
    assert history.revision() == 2
    history.add(do, undo)
    assert history.revision() == 3
    history.undo()
    assert history.revision() == 2
    history.add(do, undo)
    assert history.revision() == 4
    history.undo()
    assert history.revision() == 2
    history.redo()
    assert history.revision() == 4
    
def testSound():
    from copy import copy
    snd = Sound()

    # synthesize a sine wave
    from math import sin
    SR = 44100
    f0 = 440
    time = 5
    sine = [sin(2 * 3.14 * f0/SR * x) for x in range(time * SR)]
    start, end = 2*SR, 4*SR
    sine2 = copy(sine)
    snd.frames = copy(sine)
    del sine2[start:end]
    snd.cut(start, end)
    assert snd.frames.tolist() == sine2

    snd.history.undo()
    assert snd.frames.tolist() == sine
    snd.history.undo()
    assert snd.frames.tolist() == sine
    snd.history.redo()
    assert snd.frames.tolist() == sine2

    clip = snd.copy(start, end)
    snd.paste(0, 0, clip)
    assert snd.frames.tolist() == sine2[start:end] + sine2
    snd.history.undo()
    assert snd.frames.tolist() == sine2
    snd.history.redo()
    assert snd.frames.tolist() == sine2[start:end] + sine2

    # test with a mono file
    snd = Sound("../sounds/test1.wav")
    assert snd.frames != []
    start = 4444
    end = 55555
    data = snd.frames.tolist()
    data2 = snd.frames.tolist()
    del data2[start:end]
    snd.cut(start, end)
    assert snd.frames.tolist() == data2
    snd.history.undo()
    assert snd.frames.tolist() == data
    snd.history.undo()
    assert snd.frames.tolist() == data
    snd.history.redo()
    assert snd.frames.tolist() == data2
    clip = snd.copy(start, end)
    snd.paste(0, 0, clip)
    assert snd.frames.tolist() == data2[start:end] + data2
    snd.history.undo()
    assert snd.frames.tolist() == data2
    snd.paste(0, 0, clip)
    assert snd.frames.tolist() == data2[start:end] + data2

    # test with a stereo file
    snd = Sound("../sounds/test2.wav")
    assert snd.frames.tolist() != []
    start = 4444
    end = 55555
    data = copy(snd.frames.tolist())
    data2 = copy(snd.frames.tolist())
    del data2[start:end]
    snd.cut(start, end)
    assert snd.frames.tolist() == data2
    snd.history.undo()
    assert snd.frames.tolist() == data
    snd.history.undo()
    assert snd.frames.tolist() == data
    snd.history.redo()
    assert snd.frames.tolist() == data2
    clip = snd.copy(start, end)
    snd.paste(0, 0, clip)
    assert snd.frames.tolist() == data2[start:end] + data2
    snd.history.undo()
    assert snd.frames.tolist() == data2
    snd.paste(0, 0, clip)
    assert snd.frames.tolist() == data2[start:end] + data2

    # test save_as()
    import os
    snd = Sound("../sounds/test1.wav")
    outfile = "/tmp/test.wav"
    snd.save_as(outfile)
    assert os.path.exists(outfile)
    assert snd.filename == outfile
    snd2 = Sound(outfile)
    assert abs((snd.frames - snd2.frames).max()) < 0.0001 # quantization errors!
    os.remove(outfile)

    snd = Sound("../sounds/test2.wav")
    outfile = "/tmp/test2.wav"
    snd.save_as(outfile)
    assert os.path.exists(outfile)
    assert snd.filename == outfile
    snd2 = Sound(outfile)
    assert snd.frames.size == snd2.frames.size
    assert abs((snd.frames - snd2.frames).flatten().max()) == 0
    os.remove(outfile)

    # test mix_channels
    #
    # stereo to mono
    frames = numpy.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    out = mix_channels(frames, [[0.5, 0.5]])
    assert out.tolist() == [1, 2, 3, 4]
    #
    # mono to stereo
    frames = numpy.array([1, 2, 3, 4])
    out = mix_channels(frames, [[1], [1]])
    assert out.tolist() == [[1, 1], [2, 2], [3, 3], [4, 4]]
    #
    # channel 0 to stereo
    c0 = [1, 2, 3, 4]
    c1 = [5, 6, 7, 8]
    frames = numpy.array([c0, c1]).transpose()
    out = mix_channels(frames, [[1, 0], [1, 0]])
    assert out.tolist() == [[1, 1], [2, 2], [3, 3], [4, 4]]

    # test mix_channels_auto()
    #
    # stereo to mono
    snd = Sound()
    frames = numpy.array([[1, 1.5], [2, 2.5], [3, 3.5], [4, 4.5]])
    out = mix_channels_auto(frames, 1)
    assert out.tolist() == [1.25, 2.25, 3.25, 4.25]
    #    
    # mono to stereo
    frames = numpy.array([1, 2, 3, 4])
    out = mix_channels_auto(frames, 2)
    assert out.tolist() == [[1, 1], [2, 2], [3, 3], [4, 4]]
    #
    # 1 --> 3
    frames = numpy.array([1, 2, 3, 4])
    out = mix_channels_auto(frames, 3)
    assert out.tolist() == [[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]]
    # 
    # no change
    frames = numpy.array([1, 2, 3, 4])
    out = mix_channels_auto(frames, 1)
    assert out.tolist() == [1, 2, 3, 4]

    # test cut
    snd = Sound()
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = snd.cut(1, 3)
    assert snd.frames.tolist() == [1, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    # clip is a copy
    clip[0] = 9
    assert snd.frames.tolist() == [1, 2, 3, 4]

    # test copy
    snd = Sound()
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = snd.copy(1, 3)
    assert clip.tolist() == [2, 3]
    # clip is a copy
    clip[0] = 9
    assert snd.frames.tolist() == [1, 2, 3, 4]

    # test paste
    snd = Sound()
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = numpy.array([22, 33])
    snd.paste(1, 3, clip)
    assert snd.frames.tolist() == [1, 22, 33, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    snd.paste(1, 1, clip)
    assert snd.frames.tolist() == [1, 22, 33, 2, 3, 4]
    #
    # paste a stereo clip into a mono file
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = numpy.array([[30, 10], [40, 20]])
    snd.paste(1, 3, clip)
    assert snd.frames.tolist() == [1, 20, 30, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    snd.paste(1, 1, clip)
    assert snd.frames.tolist() == [1, 20, 30, 2, 3, 4]
    #
    # paste a mono clip into a stereo file
    snd.frames = numpy.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    clip = numpy.array([22, 33])
    snd.paste(1, 3, clip)
    assert snd.frames.tolist() == [[1, 1], [22, 22], [33, 33], [4, 4]]
    snd.undo()
    assert snd.frames.tolist() == [[1, 1], [2, 2], [3, 3], [4, 4]]
    snd.paste(1, 1, clip)
    assert snd.frames.tolist() == [[1, 1], [22, 22], [33, 33],\
                                                        [2, 2], [3, 3], [4, 4]]

    # test mix
    snd = Sound()
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = numpy.array([2, 3])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [1, 4, 6, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    #
    # clip smaller than selection
    snd = Sound()
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = numpy.array([2])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [1, 4, 3, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    #
    # must not go behind selection
    clip = numpy.array([2, 3, 4])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [1, 4, 6, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    #
    clip = numpy.array([2, 3])
    snd.mix(1, 1, clip)
    assert snd.frames.tolist() == [1, 4, 6, 4]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    #
    # mix() may extend sound length when there is no selection.
    clip = numpy.array([2, 3, 4, 5])
    snd.mix(1, 1, clip)
    assert snd.frames.tolist() == [1, 4, 6, 8, 5]
    snd.undo()
    assert snd.frames.tolist() == [1, 2, 3, 4]
    #
    # stereo too
    snd = Sound()
    snd.frames = numpy.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    clip = numpy.array([[2, 2], [3, 3]])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [[1, 1], [4, 4], [6, 6], [4, 4]]
    snd.undo()
    assert snd.frames.tolist() == [[1, 1], [2, 2], [3, 3], [4, 4]]
    #
    snd = Sound()
    snd.frames = numpy.array([[1, 1], [2, 1], [3, 2], [4, 4]])
    clip = numpy.array([[2, 2], [3, 3]])
    snd.mix(1, 1, clip) # no selection
    assert snd.frames.tolist() == [[1, 1], [4, 3], [6, 5], [4, 4]]
    snd.undo()
    assert snd.frames.tolist() == [[1, 1], [2, 1], [3, 2], [4, 4]]
    #
    # mix a stereo clip into a mono file
    snd.frames = numpy.array([1, 2, 3, 4])
    clip = numpy.array([[30, 10], [40, 20]])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [1, 22, 33, 4]
    #
    # mix a mono clip into a stereo file
    snd.frames = numpy.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    clip = numpy.array([20, 30])
    snd.mix(1, 3, clip)
    assert snd.frames.tolist() == [[1, 1], [22, 22], [33, 33], [4, 4]]

    # Do not crash when saving with None as filename
    snd = Sound()
    try:
        snd.save()
    except:
        print "OK"
    else:
        assert False
        
if __name__ == '__main__':
    testAction()
    testSound()
    testHistory()
